# coding=utf-8
"""Base diagnostic for fixing files"""
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, \
    DiagnosticDomainOption, DiagnosticVariableOption


class FixFile(Diagnostic):
    """
    Base diagnostic class for diagnostic that read and reqrite the same files

    Must be derived

    :original author: Javier Vegas-Regidor<javier.vegas@bsc.es>

    :created: July 2016

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable's name
    :type variable: str
    :param domain: variable's domain
    :type domain: ModelingRealm
    """

    def __init__(self, data_manager, startdate, member, chunk,
                 domain, variable, grid):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.domain = domain
        self.grid = grid

        self.variable_file = None
        self.corrected = None

    _STR_PREFIX = None

    def __str__(self):
        return '{0._STR_PREFIX} Startdate: {0.startdate} Member: {0.member} ' \
            'Chunk: {0.chunk} Variable: {0.domain}:{0.variable} ' \
            'Grid: {0.grid}'.format(self)

    def __eq__(self, other):
        if self._different_type(other):
            return False

        return self.startdate == other.startdate and \
            self.member == other.member and \
            self.chunk == other.chunk and \
            self.domain == other.domain and \
            self.variable == other.variable and \
            self.grid == self.grid

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: variable, domain, grid
        :type options: list[str]
        :return:
        """
        options_available = cls._get_options(diags)
        options = cls.process_options(options, options_available)
        job_list = list()
        chunk_list = diags.config.experiment.get_chunk_list()
        for startdate, member, chunk in chunk_list:
            job_list.append(
                cls(diags.data_manager, startdate, member, chunk,
                    options['domain'], options['variable'], options['grid'])
            )
        return job_list

    @classmethod
    def _get_options(cls, diags):
        return [
            DiagnosticDomainOption(),
            DiagnosticVariableOption(diags.data_manager.config.var_manager),
            DiagnosticOption('grid', '')
        ]

    def request_data(self):
        """Request data required by the diagnostic"""
        self.variable_file = self.request_chunk(
            self.domain, self.variable,
            self.startdate, self.member, self.chunk,
            grid=self.grid, to_modify=True
        )

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.corrected = self.declare_chunk(
            self.domain, self.variable,
            self.startdate, self.member, self.chunk,
            grid=self.grid
        )
