# coding=utf-8
"""Compute the sea ice extent , area and volume  in both hemispheres or a specified region"""
import os
import six

import numpy as np

import iris
import iris.analysis
import iris.coords
import iris.util
from bscearth.utils.log import Log

from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticBasinListOption, DiagnosticBoolOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.utils import Utils, TempFile


# noinspection PyUnresolvedReferences


class Sivol2d(Diagnostic):
    """
    Compute the sea ice volumen from sic and sit

    Parameters
    ----------
    data_manager: DataManager
    startdate: str
    member: int
    chunk: init
    domain: ModellingRealm
    variable: str
    basin: list of Basin
    mask: numpy.array
    """

    alias = 'sivol2d'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, var_manager, data_convention):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.generated = {}
        self.var_manager = var_manager
        self.sic_varname = self.var_manager.get_variable('sic').short_name
        self.sit_varname = self.var_manager.get_variable('sit').short_name
        self.sivol_varname = self.var_manager.get_variable('sivol').short_name
        self.data_convention = data_convention

        self.results = {}
        for var in ('siarean', 'siareas', 'siextentn', 'siextents'):
            self.results[var] = {}

    def __str__(self):
        return 'Sivol2d Startdate: {0.startdate} Member: {0.member} Chunk: {0.chunk}'.format(self)

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: basin
        :type options: list[str]
        :return:
        """
        options_available = []
        options = cls.process_options(options, options_available)

        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job_list.append(
                Sivol2d(
                    diags.data_manager, startdate, member, chunk,
                    diags.config.var_manager, diags.config.data_convention
                )
            )
        return job_list

    def request_data(self):
        """Request data required by the diagnostic"""
        self.sic = self.request_chunk(ModelingRealms.seaIce, self.sic_varname,
                                      self.startdate, self.member, self.chunk)
        self.sit = self.request_chunk(ModelingRealms.seaIce, self.sit_varname,
                                      self.startdate, self.member, self.chunk)

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.sivol = self.declare_chunk(ModelingRealms.seaIce, self.sivol_varname,
                                        self.startdate, self.member, self.chunk)

    def compute(self):
        """Run the diagnostic"""
        coordinates = ' '.join(('time', 'leadtime', 'time_centered',
                                self.data_convention.lon_name, self.data_convention.lat_name))
        handler = Utils.open_cdf(self.sic.local_file)
        handler.variables[self.sic_varname].coordinates = coordinates
        handler.close()
        sic = iris.load_cube(self.sic.local_file)
        if sic.units.origin == '%' and sic.data.max() < 2:
            sic.units = '1.0'
        sic.convert_units('1.0')

        handler = Utils.open_cdf(self.sit.local_file)
        handler.variables[self.sit_varname].coordinates = coordinates
        handler.close()
        sit = iris.load_cube(self.sit.local_file)

        sivol = sit * sic.data
        del sit
        del sic
        sivol.var_name = self.sivol_varname
        sivol.standard_name = "sea_ice_thickness"
        sivol.long_name = "Total volume of sea ice divided by grid-cell area" \
                          " (this used to be called ice thickness in CMIP5)"
        temp = TempFile.get()
        iris.save(sivol, temp, zlib=True)
        del sivol
        self.sivol.set_local_file(temp)
