# coding=utf-8
"""Diagnostic to compute regional averages"""
import iris
import iris.util
import iris.coords
import iris.analysis
import iris.exceptions
from iris.coord_categorisation import add_categorised_coord
from iris.cube import Cube, CubeList

import numpy as np
import numba

from earthdiagnostics.box import Box
from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticIntOption, DiagnosticDomainOption, \
    DiagnosticBoolOption, DiagnosticBasinOption, DiagnosticVariableOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.utils import Utils, TempFile


class ZonalMean(Diagnostic):
    """
    Computes the zonal mean value of the field (weighted).

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable to average
    :type variable: str
    :param box: box used to restrict the vertical mean
    :type box: Box
    """

    alias = 'zonmean'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, basin, grid_point):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.domain = domain
        self.variable = variable
        self.basin = basin
        self.grid_point = grid_point

        self.declared = {}

        self.lat_name = 'lat'
        self.lon_name = 'lon'

    def __eq__(self, other):
        if self._different_type(other):
            return False
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.variable == other.variable and self.grid_point == other.grid_point and self.basin == other.basin

    def __str__(self):
        return 'Zonal mean Startdate: {0.startdate} Member: {0.member} Chunk: {0.chunk} Variable: {0.variable} ' \
               'Grid point: {0.grid_point}'.format(self)

    def __hash__(self):
        return hash(str(self))

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Create a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: variable, minimum depth (level), maximum depth (level)
        :type options: list[str]
        :return:
        """
        options_available = (
            DiagnosticDomainOption(),
            DiagnosticVariableOption(diags.data_manager.config.var_manager),
            DiagnosticOption('grid_point', 'T'),
            DiagnosticBasinOption('basin', Basins().Global),
        )
        options = cls.process_options(options, options_available)

        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job = ZonalMean(diags.data_manager, startdate, member, chunk,
                            options['domain'], options['variable'], options['basin'],
                            options['grid_point'].lower())
            job_list.append(job)

        return job_list

    def request_data(self):
        """Request data required by the diagnostic"""
        self.variable_file = self.request_chunk(
            self.domain, self.variable, self.startdate, self.member, self.chunk
        )

    def declare_data_generated(self):
        """Declare data to be generated by the diagnostic"""
        self.zonal_mean = self.declare_chunk(
            ModelingRealms.ocean, self.variable + 'zonal',
            self.startdate, self.member, self.chunk, region=self.basin
        )

    def compute(self):
        """Run the diagnostic"""
        self._fix_file_metadata()
        data = self._load_data()
        self._meand_3d_variable(data)

    def _meand_3d_variable(self, data):
        e1 = self._try_load_cube(1)
        e2 = self._try_load_cube(2)
        mask = np.squeeze(Utils.get_mask(self.basin, True))
        mask = e1.data * e2.data * mask
        if len(mask.shape) == 2:
            data.add_aux_coord(
                iris.coords.AuxCoord(mask.data, long_name='mask'),
                data.coord_dims('latitude')
            )
        else:
            data.add_aux_coord(
                iris.coords.AuxCoord(mask.data, long_name='mask'),
                data.coord_dims('depth') + data.coord_dims('latitude')
            )

        @numba.njit()
        def get_zonal_mean(variable, weight, latitude):
            total = np.zeros(180, np.float64)
            weights = np.zeros(180, np.float64)
            for i in range(variable.shape[0]):
                for j in range(variable.shape[1]):
                    if weight[i, j] == 0:
                        continue
                    bin_value = int(round(latitude[i, j]) + 90)
                    weights[bin_value] += weight[i, j]
                    total[bin_value] += variable[i, j] * weight[i, j]
            return total / weights

        mean = iris.cube.CubeList()
        lat_coord = None
        for map_slice in data.slices_over('time'):
            # Force data loading
            map_slice.data
            surface_cubes = iris.cube.CubeList()
            for surface_slice in map_slice.slices_over('depth'):
                value = get_zonal_mean(
                    surface_slice.data,
                    surface_slice.coord('mask').points,
                    surface_slice.coord('latitude').points,
                )
                cube = Cube(value)
                cube.add_aux_coord(surface_slice.coord('depth'))
                if lat_coord is None:
                    lat_coord = surface_slice.coord('latitude')
                    lat_coord = lat_coord.copy(
                        np.arange(-90, 90, dtype=np.float32)
                    )
                    lat_coord = iris.coords.DimCoord.from_coord(lat_coord)
                cube.add_dim_coord(lat_coord, 0)
                surface_cubes.append(cube)
            time_cube = surface_cubes.merge_cube()
            time_cube.add_aux_coord(map_slice.coord('time'))
            mean.append(time_cube)
        cube = mean.merge_cube()
        cube.var_name = 'result'
        cube.units = data.units
        cube.attributes = data.attributes
        temp = TempFile.get()
        iris.save(cube, temp)
        self.zonal_mean.set_local_file(temp, rename_var='result', region=self.basin)

    def _try_load_cube(self, number):
        try:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}'.format(number, self.grid_point))
        except iris.exceptions.ConstraintMismatchError:
            cube = iris.load_cube('mesh_hgr.nc', 'e{0}{1}_0'.format(number, self.grid_point))
        cube = iris.util.squeeze(cube)
        dims = len(cube.shape)
        try:
            cube.coord('i')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 1]), var_name='i'), dims - 1)
        try:
            cube.coord('j')
        except iris.exceptions.CoordinateNotFoundError:
            cube.add_dim_coord(iris.coords.DimCoord(np.arange(cube.shape[dims - 2]), var_name='j'), dims - 2)
        return cube

    def _load_data(self):
        coords = []
        handler = Utils.open_cdf(self.variable_file.local_file)
        for variable in handler.variables:
            if variable in ('time', 'lev', 'lat', 'lon', 'latitude', 'longitude', 'leadtime', 'time_centered'):
                coords.append(variable)
            if variable == 'time_centered':
                handler.variables[variable].standard_name = ''

        handler.variables[self.variable].coordinates = ' '.join(coords)
        handler.close()

        data = iris.load_cube(self.variable_file.local_file)
        return self._rename_depth(data)

    def _rename_depth(self, data):
        for coord_name in ('model_level_number', 'Vertical T levels', 'lev'):
            if data.coords(coord_name):
                coord = data.coord(coord_name)
                coord.standard_name = 'depth'
                coord.long_name = 'depth'
                break
        return data

    def _fix_file_metadata(self):
        handler = Utils.open_cdf(self.variable_file.local_file)
        var = handler.variables[self.variable]
        coordinates = ''
        has_levels = False
        for dimension in handler.variables.keys():
            if dimension in ['time', 'lev', 'lat', 'latitude', 'lon', 'longitude', 'i', 'j']:
                coordinates += ' {0}'.format(dimension)
            if dimension == 'lev':
                has_levels = True
        var.coordinates = coordinates
        handler.close()
        return has_levels
