# coding=utf-8
from unittest import TestCase

from mock import Mock, patch
import os
from tempfile import mktemp

# import dummydata
import iris

from earthdiagnostics.box import Box
from earthdiagnostics.diagnostic import DiagnosticVariableOption, DiagnosticOptionError
from earthdiagnostics.frequency import Frequencies
from earthdiagnostics.general.scale import Scale
from earthdiagnostics.modelingrealm import ModelingRealms


class TestScale(TestCase):

    def setUp(self):
        self.data_manager = Mock()

        self.diags = Mock()
        self.diags.config.experiment.get_chunk_list.return_value = (('20010101', 0, 0), ('20010101', 0, 1))
        self.diags.config.experiment.startdates = ['20010101', ]
        self.diags.config.frequency = Frequencies.monthly

        self.box = Box()
        self.box.min_depth = 0
        self.box.max_depth = 100

        self.var_file = mktemp('.nc')

    def tearDown(self):
        if os.path.exists(self.var_file):
            os.remove(self.var_file)

    def fake_parse(self, value):
        return value

    @patch.object(DiagnosticVariableOption, 'parse', fake_parse)
    def test_generate_jobs(self):
        jobs = Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', '',
                                        float('nan'), float('nan'), Frequencies.monthly, False))
        self.assertEqual(jobs[1], Scale(self.data_manager, '20010101', 0, 1, 0, 0, ModelingRealms.atmos, 'var', '',
                                        float('nan'), float('nan'), Frequencies.monthly, False))

        jobs = Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0', 'grid'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        float('nan'), float('nan'), Frequencies.monthly, False))
        self.assertEqual(jobs[1], Scale(self.data_manager, '20010101', 0, 1, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        float('nan'), float('nan'), Frequencies.monthly, False))

        jobs = Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0', 'grid', '0', '100'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.monthly, False))
        self.assertEqual(jobs[1], Scale(self.data_manager, '20010101', 0, 1, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.monthly, False))

        jobs = Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0', 'grid', '0', '100', '3hr'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.three_hourly, False))
        self.assertEqual(jobs[1], Scale(self.data_manager, '20010101', 0, 1, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.three_hourly, False))

        jobs = Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0', 'grid', '0', '100', '3hr',
                                                True])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.three_hourly, True))
        self.assertEqual(jobs[1], Scale(self.data_manager, '20010101', 0, 1, 0, 0, ModelingRealms.atmos, 'var', 'grid',
                                        0, 100, Frequencies.three_hourly, True))

        with self.assertRaises(DiagnosticOptionError):
            Scale.generate_jobs(self.diags, ['diagnostic'])

        with self.assertRaises(DiagnosticOptionError):
            Scale.generate_jobs(self.diags, ['diagnostic', 'atmos', 'var', '0', '0', 'grid', '0', '100', '3hr', 'True',
                                             'extra'])

    def test_str(self):
        scale = Scale(self.data_manager, '20010101', 0, 0, 0, 0, ModelingRealms.atmos, 'var', 'grid', 0, 100,
                      Frequencies.three_hourly, False)
        self.assertEqual(str(scale),
                         'Scale output Startdate: 20010101 Member: 0 Chunk: 0 Scale value: 0 Offset: 0 '
                         'Variable: atmos:var Frequency: 3hr Apply mask: False')

    # def test_compute_factor(self):

    #     scale = Scale(self.data_manager, '20010101', 0, 0, 10, 0, ModelingRealms.atmos, 'ta', 'grid', 1, 100,
    #                   Frequencies.three_hourly, False)
    #     cube = self._get_data_and_test(scale)
    #     self.assertEqual(cube.data.max(), 10)

    # def test_compute_offset(self):
    #     scale = Scale(self.data_manager, '20010101', 0, 0, 1, 10, ModelingRealms.atmos, 'ta', 'grid', 1, 100,
    #                   Frequencies.three_hourly, False)
    #     cube = self._get_data_and_test(scale)
    #     self.assertEqual(cube.data.max(), 11)

    # def test_compute_too_low(self):
    #     scale = Scale(self.data_manager, '20010101', 0, 0, 0, 10, ModelingRealms.atmos, 'ta', 'grid', 10, 100,
    #                   Frequencies.three_hourly, False)
    #     cube = self._get_data_and_test(scale)
    #     self.assertEqual(cube.data.max(), 1)

    # def test_compute_too_high(self):
    #     scale = Scale(self.data_manager, '20010101', 0, 0, 0, 10, ModelingRealms.atmos, 'ta', 'grid', 0, 0.5,
    #                   Frequencies.three_hourly, False)
    #     cube = self._get_data_and_test(scale)
    #     self.assertEqual(cube.data.max(), 1)

    # def _get_data_and_test(self, scale):
    #     dummydata.model2.Model2(oname=self.var_file, var='ta', start_year=2000, stop_year=2000, method='constant',
    #                             constant=1)
    #     scale.variable_file = Mock()
    #     scale.variable_file.local_file = self.var_file
    #     scale.corrected = Mock()
    #     scale.compute()
    #     scale.corrected.set_local_file.assert_called_once()
    #     cube = iris.load_cube(scale.corrected.set_local_file.call_args[0][0])
    #     self.assertEqual(cube.data.max(), cube.data.min())
    #     return cube
