# coding=utf-8
from unittest import TestCase

from mock import Mock, patch

from earthdiagnostics.box import Box
from earthdiagnostics.constants import Basins
from earthdiagnostics.diagnostic import DiagnosticOptionError, DiagnosticVariableOption
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.ocean.regionmean import RegionMean
from earthdiagnostics.utils import TempFile


class TestRegionMean(TestCase):

    def setUp(self):
        self.data_manager = Mock()
        self.diags = Mock()
        self.diags.config.experiment.get_chunk_list.return_value = (('20010101', 0, 0), ('20010101', 0, 1))

    def fake_parse(self, value):
        if not value:
            raise DiagnosticOptionError
        return value

    @staticmethod
    def fake_get():
        return 'tempfile'

    @patch.object(DiagnosticVariableOption, 'parse', fake_parse)
    @patch.object(TempFile, 'get', fake_get)
    def test_generate_jobs(self):

        box = Box()
        box.min_depth = -1
        box.max_depth = -1

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))

        box = Box()
        box.min_depth = 1.0
        box.max_depth = 10.0

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global', '1', '10'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, True, 'weights', False, Basins().Global))

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global', '1', '10', 'false'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, False, ''))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, False, ''))

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global', '1', '10', 'false',
                                                     'True'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, True, ''))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, True, ''))

        jobs = RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global', '1', '10', 'false',
                                                     'True', 'grid'])
        self.assertEqual(len(jobs), 2)
        self.assertEqual(jobs[0], RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, True, 'grid'))
        self.assertEqual(jobs[1], RegionMean(self.data_manager, '20010101', 0, 1, ModelingRealms.ocean, 'var',
                                             box, False, Basins().Global, True, 'grid'))

        with self.assertRaises(DiagnosticOptionError):
            RegionMean.generate_jobs(self.diags, ['diagnostic'])

        with self.assertRaises(DiagnosticOptionError):
            RegionMean.generate_jobs(self.diags, ['diagnostic', 'ocean', 'var', 'U', 'global', '1', '10', 'false',
                                                  'True', 'grid', 'extra'])

    def test_str(self):
        box = Box()
        box.min_depth = 1
        box.max_depth = 10
        diag = RegionMean(self.data_manager, '20010101', 0, 0, ModelingRealms.ocean, 'var', box, False,
                          True, Basins().Global, 'T')
        self.assertEqual(str(diag),
                         'Region mean Startdate: 20010101 Member: 0 Chunk: 0 Variable: var Box: 1-10 '
                         'Save 3D: False Save variance: True Grid point: T')
