# coding=utf-8
import os
import shutil
import tempfile
from unittest import TestCase

import mock
from mock import Mock, create_autospec

from earthdiagnostics.cmormanager import CMORManager, MergeYear
from earthdiagnostics.modelingrealm import ModelingRealms
from earthdiagnostics.data_convention import DataConvention


class TestCMORManager(TestCase):

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        os.mkdir(os.path.join(self.tmp_dir, 'expid'))

        self.convention = create_autospec(DataConvention)
        self.convention.name = 'specs'
        self.convention.lat_name = 'lat'
        self.convention.lon_name = 'lon'
        self.convention.get_startdate_path.return_value = os.path.join(self.tmp_dir, 'expid', 'startdate')
        self.convention.get_cmor_folder_path.return_value = os.path.join(self.tmp_dir, 'expid', 'startdate', 'cmor')
        self.convention.get_file_name.return_value = 'filename'
        self.convention.get_file_path.return_value = os.path.join(self.tmp_dir, 'expid', 'startdate', 'cmor',
                                                                  'filename')
        self.convention.is_cmorized.return_value = True

        self.config = Mock()
        self.config.data_dir = self.tmp_dir
        self.config.scratch_dir = os.path.join(self.tmp_dir, 'scratch')
        self.config.data_convention = self.convention
        self.config.frequency = '6hr'
        self.config.data_type = 'exp'
        self.config.experiment.expid = 'expid'
        self.config.experiment.model = 'model'
        self.config.experiment.experiment_name = 'expname'
        self.config.experiment.institute = 'institute'
        self.config.experiment.member_count_start = 0
        self.config.experiment.chunk_size = 12
        self.config.experiment.num_chunks = 1
        self.config.experiment.calendar = 'standard'
        self.config.experiment.atmos_timestep = 3
        self.config.experiment.ocean_timestep = 6
        self.config.experiment.get_member_str.return_value = 'r1i1p1'

        self.config.cmor.initialization_number = 1
        self.config.cmor.version = ''
        self.config.cmor.default_ocean_grid = 'ocean_grid'
        self.config.cmor.default_atmos_grid = 'atmos_grid'
        self.config.cmor.activity = 'activity'
        self.config.cmor.force = False
        self.config.cmor.force_untar = False
        self.config.cmor.cmorize_atmosphere = True
        self.config.cmor.cmorize_ocean = True
        self.config.cmor.skip_prepare = False
        self.config.cmor.append_startdate = False

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_find_data(self):
        cmor_manager = CMORManager(self.config)
        self.assertEqual(cmor_manager.cmor_path, os.path.join(self.tmp_dir, 'expid'))

    def test_find_data_fail(self):
        os.rmdir(os.path.join(self.tmp_dir, self.config.experiment.expid))
        with self.assertRaises(Exception):
            CMORManager(self.config)

    def test_find_data_with_model(self):
        os.makedirs(os.path.join(self.tmp_dir, 'model', self.config.experiment.expid))
        os.rmdir(os.path.join(self.tmp_dir, self.config.experiment.expid))
        cmor_manager = CMORManager(self.config)
        self.assertEqual(cmor_manager.cmor_path, os.path.join(self.tmp_dir, 'model', 'expid'))

    def test_find_data_with_ecearth_fix(self):
        self.config.experiment.model = 'EC-Earth'
        os.makedirs(os.path.join(self.tmp_dir, 'ecearth', self.config.experiment.expid))
        os.rmdir(os.path.join(self.tmp_dir, self.config.experiment.expid))
        cmor_manager = CMORManager(self.config)
        self.assertEqual(cmor_manager.cmor_path, os.path.join(self.tmp_dir, 'ecearth', 'expid'))

    def test_find_data_with_type_and_model(self):
        os.makedirs(os.path.join(self.tmp_dir, 'exp', 'model', self.config.experiment.expid))
        os.rmdir(os.path.join(self.tmp_dir, self.config.experiment.expid))
        cmor_manager = CMORManager(self.config)
        self.assertEqual(cmor_manager.cmor_path, os.path.join(self.tmp_dir, 'exp', 'model', 'expid'))

    def test_file_exists(self):
        with mock.patch('os.path.isfile') as isfile:
            cmor_manager = CMORManager(self.config)
            isfile.return_value = True
            self.assertTrue(cmor_manager.file_exists(ModelingRealms.ocean, 'var', '20011101', 1, 1))
            isfile.return_value = False
            self.assertFalse(cmor_manager.file_exists(ModelingRealms.ocean, 'var', '20011101', 1, 1))

    def test_file_exists_multiple_versions(self):
        with mock.patch('os.path.isfile') as isfile:
            cmor_manager = CMORManager(self.config)
            isfile.return_value = True
            self.assertTrue(cmor_manager.file_exists(ModelingRealms.ocean, 'var', '20011101', 1, 1,
                                                     possible_versions=('version1', 'version2')))
            isfile.return_value = False
            self.assertFalse(cmor_manager.file_exists(ModelingRealms.ocean, 'var', '20011101', 1, 1,
                                                      possible_versions=('version1', 'version2')))

    @mock.patch('earthdiagnostics.cmormanager.Cmorizer', autospec=True)
    def test_prepare_cmorize(self, mock_cmor):
        mock_instance = mock_cmor.return_value
        self.convention.is_cmorized.return_value = False
        cmor_manager = CMORManager(self.config)
        self.config.experiment.get_member_list.return_value = (('20000101', 2),)
        cmor_manager.prepare()
        mock_instance.cmorize_ocean.assert_called_once()
        mock_instance.cmorize_atmos.assert_called_once()

    @mock.patch('earthdiagnostics.cmormanager.Cmorizer', autospec=True)
    def test_prepare_cmorize_force(self, mock_cmor):
        mock_instance = mock_cmor.return_value
        self.config.cmor.force = True
        cmor_manager = CMORManager(self.config)
        self.config.experiment.get_member_list.return_value = (('20000101', 2),)
        cmor_manager.prepare()
        mock_instance.cmorize_ocean.assert_called_once()
        mock_instance.cmorize_atmos.assert_called_once()

    @mock.patch('earthdiagnostics.cmormanager.CMORManager.is_cmorized')
    @mock.patch('earthdiagnostics.utils.Utils.unzip')
    @mock.patch('earthdiagnostics.utils.Utils.untar')
    def test_prepare_cmorize_force_untar(self, mock_untar, mock_unzip, mock_cmorized):
        original_cmor_path = os.path.join(self.config.data_dir, self.config.experiment.expid,
                                          'original_files', 'cmorfiles')
        os.makedirs(original_cmor_path)
        self.config.experiment.get_chunk_start_str.return_value = '20000101'
        cmor_prefix = 'CMORT_{0}_{1}_{2}_{3}-'.format(self.config.experiment.expid, '20000101', 'r1i1p1',
                                                      '20000101')
        tempfile.mkstemp('.tar.gz', cmor_prefix, original_cmor_path)
        tempfile.mkstemp('.tar', cmor_prefix, original_cmor_path)
        mock_cmorized.return_value = True
        self.config.cmor.force_untar = True
        cmor_manager = CMORManager(self.config)
        self.config.experiment.get_member_list.return_value = (('20000101', 2),)
        cmor_manager.prepare()
        self.convention.create_links.assert_called_once()
        mock_unzip.assert_called_once()
        mock_untar.assert_called_once()

    @mock.patch('earthdiagnostics.cmormanager.CMORManager.is_cmorized')
    def test_prepare_already_cmorized(self, mock_cmorized):
        mock_cmorized.return_value = True
        self.config.cmor.force_untar = False
        cmor_manager = CMORManager(self.config)
        self.config.experiment.get_member_list.return_value = (('20000101', 2),)
        cmor_manager.prepare()

    @mock.patch('earthdiagnostics.cmormanager.CMORManager.is_cmorized')
    def test_prepare_cmorization_not_requested(self, mock_cmorized):
        mock_cmorized.return_value = False
        self.config.cmor.chunk_cmorization_requested.return_value = False
        cmor_manager = CMORManager(self.config)
        self.config.experiment.get_member_list.return_value = (('20000101', 2),)
        cmor_manager.prepare()

    def test_prepare_meteofrance(self):
        self.convention.name = 'meteofrance'
        cmor_manager = CMORManager(self.config)
        cmor_manager.prepare()

    def test_is_not_cmorized(self):
        self.convention.is_cmorized.return_value = False
        cmor_manager = CMORManager(self.config)
        self.assertFalse(cmor_manager.is_cmorized('20000101', 1, 1, ModelingRealms.ocean))

    def test_link_file(self):
        cmor_manager = CMORManager(self.config)
        cmor_var = Mock()
        cmor_manager.link_file(ModelingRealms.ocean, 'var', cmor_var, '20010101', 1, 1)
        self.convention.create_link.assert_called_once()

    def test_request_chunk(self):
        cmor_manager = CMORManager(self.config)
        datafile = cmor_manager.request_chunk(ModelingRealms.ocean, 'var', '20010101', 1, 1)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

    def _mock_filepath(self):
        return os.path.join(self.tmp_dir, 'expid/startdate/cmor/filename')

    def test_declare_chunk(self):
        cmor_manager = CMORManager(self.config)
        datafile = cmor_manager.declare_chunk(ModelingRealms.ocean, 'var', '20010101', 1, 1)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

        mock_frequency = Mock()
        datafile = cmor_manager.declare_chunk(ModelingRealms.ocean, 'var', '20010101', 1, 1,
                                              frequency=mock_frequency)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

        self.config.var_manager.get_variable.return_value = None
        datafile = cmor_manager.declare_chunk(ModelingRealms.ocean, 'var', '20010101', 1, 1)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

    def test_request_year(self):
        self.config.experiment.get_year_chunks.return_value = (1, 2)
        mock_diagnostic = Mock()
        cmor_manager = CMORManager(self.config)
        datafile = cmor_manager.request_year(mock_diagnostic, ModelingRealms.ocean, 'var', '20010101', 1, 2000)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

    def test_declare_year(self):
        cmor_manager = CMORManager(self.config)
        datafile = cmor_manager.declare_year(ModelingRealms.ocean, 'var', '20010101', 1, 2001)
        self.assertEqual(datafile.remote_file, self._mock_filepath())

        self.config.var_manager.get_variable.return_value = None
        datafile = cmor_manager.declare_year(ModelingRealms.ocean, 'var', '20010101', 1, 2001)
        self.assertEqual(datafile.remote_file, self._mock_filepath())


class TestMergeYear(TestCase):

    def setUp(self):
        self.data_manager = Mock()

    def test_generate_jobs(self):
        self.assertIsNone(MergeYear.generate_jobs(None, None))

    def test_eq(self):
        self.assertEqual(MergeYear(self.data_manager, ModelingRealms.ocean, 'var', 'startdate', 1, 1998),
                         MergeYear(self.data_manager, ModelingRealms.ocean, 'var', 'startdate', 1, 1998))
        self.assertEqual(MergeYear(self.data_manager, ModelingRealms.ocean, 'var', 'startdate', 1, 1998, 'grid'),
                         MergeYear(self.data_manager, ModelingRealms.ocean, 'var', 'startdate', 1, 1998, 'grid'))
