# coding=utf-8
from unittest import TestCase
from mock import Mock

from earthdiagnostics.frequency import Frequencies
from earthdiagnostics.modelingrealm import ModelingRealm, ModelingRealms


class TestModellingRealms(TestCase):

    def test_parse(self):
        self.assertEqual(ModelingRealms.parse('atmos'), ModelingRealms.atmos)
        self.assertEqual(ModelingRealms.parse('atmoschem'), ModelingRealms.atmosChem)
        self.assertEqual(ModelingRealms.parse('atmoSChem'), ModelingRealms.atmosChem)
        self.assertIsNone(ModelingRealms.parse(''))
        with self.assertRaises(ValueError):
            ModelingRealms.parse('badrealm')


class TestModellingRealm(TestCase):

    def setUp(self):
        self.realm = ModelingRealm('ocean')
        self.data_convention = Mock()
        self.data_convention.name = 'specs'

    def test__repr__(self):
        self.assertEqual(str(self.realm), repr(self.realm))

    def test_constructor_fail_on_bad_realm(self):
        with self.assertRaises(ValueError):
            ModelingRealm('badrealm')

    def test_comparison(self):
        self.assertEqual(ModelingRealm('ocean'), self.realm)
        self.assertNotEqual(ModelingRealm('OCEAN'), self.realm)
        self.assertNotEqual(ModelingRealm('atmos'), self.realm)

    def test_get_omon(self):
        self.assertEqual(self.realm.get_table_name(Frequencies.monthly, 'specs'), 'Omon')

    def test_get_oimon(self):
        self.assertEqual(ModelingRealm('seaIce').get_table_name(Frequencies.monthly, self.data_convention), 'OImon')

    def test_get_simon(self):
        self.data_convention.name = 'cmip6'
        self.assertEqual(ModelingRealm('seaIce').get_table_name(Frequencies.monthly, self.data_convention), 'SImon')

    def test_get_limon(self):
        self.assertEqual(ModelingRealm('landIce').get_table_name(Frequencies.monthly, self.data_convention), 'LImon')

    def test_get_day(self):
        self.assertEqual(ModelingRealm('atmos').get_table_name(Frequencies.daily, self.data_convention), 'day')

    def test_get_6hrplev(self):
        self.assertEqual(ModelingRealm('atmos').get_table_name(Frequencies.six_hourly, self.data_convention), '6hrPlev')

    def test_get_varfolder(self):
        self.assertEqual(ModelingRealm('ocean').get_varfolder('var', 6, 3), 'var_f6h')
        self.assertEqual(ModelingRealm('atmos').get_varfolder('var', 6, 3), 'var_f3h')
        frequency = Mock()
        frequency.__str__ = Mock()
        frequency.__str__.return_value = '3hr'
        frequency.frequency = '3hr'
        self.assertEqual(ModelingRealm('atmos').get_varfolder('var', 6, 3, frequency=frequency),
                         'var')
